"""Test functions from annotate."""
# flake8: noqa

#@pytest.mark.parametrize(
#    "start,end,read_length,expected",
#    [
#        (0, 1, 9, "concat001:0:1"),
#        (1, 9, 10, "concat001:01:09"),
#        (1, 10, 10, "concat001:01:10"),
#        # (0, 100, 1000,  "concat001:000"),
#        # (99, 100,10, "concat001:099"),
#    ],
#)
#def test_monomer_id(start, end, read_length, expected):
#    """Test monomer id."""
#    coords = ConcatemerCoords(
#        start=start, end=end,
#        read_length=read_length, subread_idx=0, subread_total=1
#    )
#    res = MonomerReadSeq.generate_id("concat001", coords)
#    assert res == expected
#
#
#def test_group_monomer_aligns(monomer_read_seqs):
#    """Test group monomer aligns."""
#    num_concatemers, aligns_per_concatemer = 2, 10
#    concatemer_ids = set()
#    for concat_id, monomers in group_aligns_by_concatemers(monomer_read_seqs):
#        assert len(list(monomers)) == aligns_per_concatemer
#        concatemer_ids.add(concat_id)
#    assert len(concatemer_ids) == num_concatemers
#    shuffled = sorted(
#        monomer_read_seqs, key=lambda x: x.read_seq.align_info.ref_name)
#    with pytest.raises(ValueError):
#        for concat_id, monomer_read_seqs in group_aligns_by_concatemers(shuffled): # noqa
#            pass
#
#
#def test_annotate_monomer_aligns(monomer_read_seqs):
#    """Test annotate monomer aligns."""
#    res = {
#        concat_id: aligns
#        for (concat_id, aligns) in annotate_monomer_alignments(monomer_read_seqs)   # noqa
#    }
#    assert len(res) == 2
#    for aligns in res.values():
#        assert len(aligns) == 10
#        for a in aligns:
#            a._update_tags()
#            tags = a.read_seq.tags
#            assert MOLECULE_TAG in tags
#            assert WALK_TAG in tags
#            assert CONCATEMER_TAG in tags
#
#
#def test_mods(mock_reads):
#    """Test mods."""
#    test_read = mock_reads["read_w_tags"]
#    header = AlignmentHeader.from_dict(
#        {"RG": [{"ID": "RG1", "SM": "sample01", "LB": "lib01"}]}
#    )
#    seg = test_read.to_align(header)
#    assert seg.modified_bases[("C", 0, "m")] == [(2, 122), (10, 128)]
#    # assert seg.get_tag("RG") == "RG1"
#
#
#def test_split_read(concatemer_unaligned: AlignedSegment):
#    """Test split read."""
#    concatemer = ConcatemerReadSeq.from_align(
#        concatemer_unaligned, as_unaligned=True, init_mod_bases=True
#    )
#    monomers = concatemer.split([5])
#    for idx, s in [(0, slice(None, 5)), (1, slice(5, None))]:
#        assert (
#            monomers[idx].read_seq.sequence
#            == concatemer_unaligned.seq[s]  # type: ignore
#        )
#        assert (
#            monomers[idx].read_seq.quality
#            == concatemer_unaligned.qual[s]  # type: ignore
#        )
#    assert monomers[0].to_align().modified_bases == {  # type: ignore
#        ("C", 0, "m"): [(2, 122)]
#    }
#    assert monomers[1].to_align().modified_bases == {  # type: ignore
#        ("C", 0, "m"): [(5, 128)]
#    }


"""Command-line interface."""
import argparse
import contextlib
import itertools
import logging
from pathlib import Path
import sys
import time

import pysam

from pore_c_py import align_tools, annotate, digest, utils, writers


def porec_parser():
    """Create CLI parser."""
    from pore_c_py import __version__
    parser = argparse.ArgumentParser(
        "pore-c-py", parents=[utils.log_level()],
        description="Tools for processing Pore-C data.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command')
    subparsers.required = True

    parser.add_argument(
        '--version', action='version',
        version='%(prog)s {}'.format(__version__))

    digest_parse = subparsers.add_parser(
        'digest',
        description=(
            "Digest concatemer sequences into monomers "
            "using a restriction enzyme."),
        help=(
            "Resulting BAM file has query names of the format "
            "<original_read_id>:<subread_start>:<subread_end> with "
            "coordinates left padded with zeroes such that they are "
            f"sortable as strings. The f{utils.CONCATEMER_ID_TAG} SAM tag "
            "is set to the original "
            f"read/concatemer ID. The f{utils.MONOMER_DATA_TAG} tag "
            "contains the relative "
            "coordinates of the monomer on the concatemer. "
            f"{utils.MONOMER_DATA_TAG}:B:i,start,end,concatemer_length,subread_idx,subread_total "  # noqa:E501
            "If modified base tags are found in the input then they will "
            "processed so that each subread has the correct ML/MM tags. "
            "Note that the 'mv' SAM tag is not handled in the same way "
            "and is removed."),
        parents=[utils.log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    digest_parse.set_defaults(func=digest_bam)
    digest_parse.add_argument(
        "input", type=Path,
        help="An unaligned BAM file or FASTQ file or a directory of same")
    digest_parse.add_argument(
        "enzyme",
        help="A restriction enzyme name eg. NlaIII")
    digest_parse.add_argument(
        "--output", type=Path, default=sys.stdout,
        help="An unaligned BAM file with a separate record for each monomer,"
        "default will write to standard out.")
    digest_parse.add_argument(
        "--recursive", default=False, action="store_true",
        help=(
            "If INPUT is a directory search recursively "
            "for files matching 'glob'")),
    digest_parse.add_argument(
        "--glob", default="*.bam",
        help="If INPUT is a directory use this glob to search for files")
    digest_parse.add_argument(
        "--max_reads", type=int, default=0,  # should be inf?
        help="Take the first n reads (useful for testing)")
    digest_parse.add_argument(
        "--remove_tags", nargs="+", default=list(),
        help="Optionally remove SAM tags from input")
    digest_parse.add_argument(
        "--threads", default=1, type=int,
        help="Compute threads of bam compression.")

    annotate_parse = subparsers.add_parser(
        "annotate",
        description=(
            "Filter a BAM of concatemer-grouped monomer alignments, "
            "and add 'walk' tag."),
        parents=[utils.log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    annotate_parse.set_defaults(func=annotate_bam)
    annotate_parse.add_argument(
        "bam", type=Path,
        help=(
            "A (concatemer) namesorted BAM of aligned "
            "and unaligned monomers."))
    annotate_parse.add_argument(
        "output_prefix", type=Path,
        help="Output files will share this prefix")
    annotate_parse.add_argument(
        "--stdout", default=False,
        help="If True output name sorted bam to stdout")
    annotate_parse.add_argument(
        "--threads", default=1, type=int,
        help="Compute threads of bam compression.")
    annotate_parse.add_argument(
        "--force", default=False, action="store_true",
        help="Overwite any existing files")
    annotate_parse.add_argument(
        "--monomers", default=False, action="store_true",
        help="Create a namesorted BAM file with Pore-C annotations")
    annotate_parse.add_argument(
        "--chromunity", default=False, action="store_true",
        help="Create a chromunity-compatible parquet")
    annotate_parse.add_argument(
        "--chromunity_merge_distance", type=int,
        help=(
            "Merge co-linear monomers that are separated by less than this"
            "distance into a single monomer"))
    annotate_parse.add_argument(
        "--summary", default=False, action="store_true")
    annotate_parse.add_argument(
        "--direct_only", default=False, action="store_true",
        help=(
            "Only output monomer pairs that are adjacent on the concatemer"
            ", don't do combinatorial expansion."))
    annotate_parse.add_argument(
        "--paired_end", default=False, action="store_true",
        help="Create a mock paired-end BAM")
    annotate_parse.add_argument(
        "--filter_pairs", action="store_true",
        help=(
            "Filter out pairs using min and max distance"))
    annotate_parse.add_argument(
        "--paired_end_minimum_distance", default=0, type=int,
        help=(
            "Filter out any pairs shorter than this distance."
            "Note setting this removes all trans pairs."))
    annotate_parse.add_argument(
        "--paired_end_maximum_distance", default=float("inf"), type=int,
        help=(
            "Filter out any pairs longer than this distance. "
            "Note setting this removes all trans pairs."))
    annotate_parse.add_argument(
        "--allow_singletons", default=False, action="store_true",
        help=(
            "If filter pairs is true, allow singletons in paired end bam"))
    annotate_parse.add_argument(
        "--allow_improper", default=False, action="store_true",
        help=(
            "If filter pairs is true, allow improper pairs in paired end bam"))
    annotate_parse.add_argument(
        "--allow_unmapped", default=False, action="store_true",
        help=(
            "If filter pairs is true, allow unmapped in paired end bam"))

    chunk_bam_parse = subparsers.add_parser(
        "chunk-bam", help="Chunk a BAM into pieces.",
        parents=[utils.log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    chunk_bam_parse.set_defaults(func=chunk_bam)
    chunk_bam_parse.add_argument(
        "input", type=Path,
        help="An unaligned BAM file or directory of same")
    chunk_bam_parse.add_argument(
        "output_prefix", type=Path,
        help="An unaligned BAM file with a separate record for each monomer")
    chunk_bam_parse.add_argument(
        "--chunk_size", type=int, default=8000,
        help="The number of reads in each chunk"),
    chunk_bam_parse.add_argument(
        "--recursive", default=False, action="store_true",
        help=(
            "If INPUT is a directory search recursively for "
            "files matching 'glob'"))
    chunk_bam_parse.add_argument(
        "--glob", default="*.fastq",
        help="If INPUT is a directory use this glob to search for files")
    chunk_bam_parse.add_argument(
        "--max_reads", type=int,
        help="Take the first n reads (useful for testing)")

    return parser


def digest_bam(args):
    """Digest entry point."""
    logger = utils.get_named_logger("Digest")
    logger.info(f"Digesting concatemers from {args.input}.")
    input_files = list(
        utils.find_files(
            args.input, glob=args.glob, recursive=args.recursive))
    with pysam.AlignmentFile(
            str(input_files[0]), "r", check_sq=False) as inputfile:
        header = align_tools.update_header(inputfile.header)
    mode = "wb"
    if (args.output is not sys.stdout) and \
            (not utils.stdout_is_regular_file()):
        mode = "wb0"
    with pysam.AlignmentFile(
            args.output, header=header,
            threads=args.threads, mode=mode) as outbam:
        for input_file in input_files:
            with pysam.AlignmentFile(
                    str(input_file), "r", check_sq=False) as inputfile:
                for monomer in digest.get_concatemer_seqs(
                    inputfile, enzyme=args.enzyme,
                        remove_tags=args.remove_tags):
                    outbam.write(monomer)
    logger.info("Finished digestion.")


def annotate_bam(args):
    """Parse BAM entry point."""
    logger = utils.get_named_logger("AnntateBAM")
    logger.info(f"Processing reads from {args.bam}")
    if all(
        x is False for x in (
            args.monomers, args.chromunity, args.summary, args.paired_end)):
        logger.error(
            "You must specify at least one of --monomers, --chromunity, "
            "--summary, or --paired_end")
        sys.exit(1)
    mode = "wb"
    if args.stdout:
        namesorted_bam = sys.stdout
        if not utils.stdout_is_regular_file():
            mode = "wb0"
    else:
        namesorted_bam = Path(f"{args.output_prefix}.ns.bam")
    paired_end_bam = Path(f"{args.output_prefix}.pe.bam")
    summary_json = Path(f"{args.output_prefix}.summary.json")
    chromunity_parquet = Path(f"{args.output_prefix}.chromunity.parquet")
    outputs = (
        namesorted_bam, paired_end_bam, summary_json, chromunity_parquet)

    if any(x.exists() for x in outputs) and not args.force:
        logger.error(
            "Some of the outputs exist, please remove before continuing.")
        sys.exit(1)

    # TODO: other outputs
    with contextlib.ExitStack() as manager:
        inbam = pysam.AlignmentFile(args.bam, "r", check_sq=False)
        manager.enter_context(inbam)

        if args.monomers:
            header = align_tools.update_header(inbam.header)
            outbam = pysam.AlignmentFile(
                namesorted_bam, mode=mode, header=header, threads=args.threads)
            manager.enter_context(outbam)
        if args.chromunity:
            chrom_writer = writers.ChromunityWriter(
                chromunity_parquet,
                merge_distance=args.chromunity_merge_distance)
            manager.enter_context(chrom_writer)
        if args.paired_end:
            pairs_bam = pysam.AlignmentFile(
                paired_end_bam, "wb", header=header, threads=args.threads)
            manager.enter_context(pairs_bam)
        if args.summary:
            summary = writers.StatsWriter(summary_json)
            manager.enter_context(summary)

        for concat_walk in annotate.annotate_alignments(inbam):
            if args.monomers:
                for aln in concat_walk:
                    outbam.write(aln)
            if args.chromunity:
                chrom_writer.write(concat_walk)
            if args.summary or args.paired_end:
                pairs = align_tools.get_pairs(concat_walk, args.direct_only)
                if args.filter_pairs:
                    pairs = align_tools.filter_pairs(
                        pairs,
                        allow_singletons=args.allow_singletons,
                        allow_improper=args.allow_improper,
                        allow_unmapped=args.allow_unmapped)
                for p in pairs:
                    if args.paired_end:
                        pairs_bam.write(p.left)
                        if p.right is not None:
                            pairs_bam.write(p.right)
                    if args.summary:
                        summary.append(p)
    logger.info("Finished BAM parsing.")


def chunk_bam(args):
    """Create chunked ubam."""
    logger = utils.get_named_logger("ChunkedBAM")
    input_files = list(utils.find_files(
        args.input, glob=args.glob, recursive=args.recursive))
    if len(input_files) == 0:
        logger.error("No input files found at {args.input}")
        sys.exit(1)

    if args.output_prefix.is_dir():
        args.output_prefix = args.output_prefix / "chunked"

    with pysam.AlignmentFile(
            str(input_files[0]), "r", check_sq=False) as inbam:
        header = align_tools.update_header(inbam.header)

    read_stream = itertools.chain(
        *(pysam.AlignmentFile(str(x), check_sq=False) for x in input_files))
    if args.max_reads is not None:
        read_stream = itertools.islice(read_stream, args.max_reads)

    def _new_file(index, bam):
        if bam is not None:
            bam.close()
        return index + 1, pysam.AlignmentFile(
            str(args.output_prefix.with_suffix(f".batch_{index}.bam")),
            mode="wb", header=header)

    def _log_time(start, index):
        elapsed = time.perf_counter() - start
        reads_per_minute = int(args.chunk_size * 60 / elapsed)
        logger.info(
            f"Wrote batch {index}: {reads_per_minute} reads/min.")
        return time.perf_counter()

    bam = None
    index, written = 0, 0
    start = time.perf_counter()
    for align in read_stream:
        if written == 0:
            index, bam = _new_file(index, bam)
        bam.write(align)
        written += 1
        if written % args.chunk_size == 0:
            written = 0
            start = _log_time(start, index)
    if written != 0:
        _log_time(start, index)
    if bam is not None:
        bam.close()
    logger.info("Finished repacking BAMs.")


def run_main():
    """Entry point."""
    parser = porec_parser()
    args = parser.parse_args()

    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s',
        datefmt='%H:%M:%S', level=logging.INFO)
    logger = logging.getLogger(__package__)
    logger.setLevel(args.log_level)
    if args.logfile:
        fh = logging.FileHandler(args.logfile, "w")
        fh.setLevel(args.log_level)
        logger.addHandler(fh)

    args.func(args)



"""Digestion of unaligned concatemers."""
import copy

from Bio import Restriction
from Bio.Seq import Seq

from pore_c_py import utils

logger = utils.get_named_logger("Digest")


def get_subread(align, start, end):
    """Get sub read."""
    seq = align.query_sequence[start:end]
    if align.query_qualities:
        qual = align.query_qualities[start:end]
    else:
        qual = None
    if align.modified_bases:
        mm_str, ml_str = "", ""
        base_indices = {}
        for mod_key, mod_data in align.modified_bases.items():
            # find the modifications that overlap the subread
            idx = [
                x for x in range(len(mod_data)) if
                start <= mod_data[x][0] < end]
            if not idx:  # no mods in this subread
                continue
            try:
                canonical_base, strand, skip_scheme, mod_type = mod_key
            except ValueError:
                canonical_base, strand, mod_type = mod_key
                skip_scheme = ""

            canonical_base = mod_key[0]
            # find the positions of that canonical base in the subread
            if canonical_base not in base_indices:
                base_indices[canonical_base] = [
                    x for x, b in enumerate(seq) if b.upper() == canonical_base
                ]

            base_offsets, probs = zip(*[mod_data[_] for _ in idx])
            deltas = []
            counter = 0
            for seq_idx in base_indices[canonical_base]:
                orig_idx = seq_idx + start
                if orig_idx in base_offsets:  # is modified
                    deltas.append(str(counter))
                    counter = 0
                else:
                    counter += 1
            assert len(deltas) == len(probs)
            prob_str = ",".join(map(str, probs))
            strand = "+" if strand == 0 else "-"
            mm_str += (
                f"{canonical_base}{strand}{mod_type}{skip_scheme}"
                f",{','.join(deltas)};"
            )
            ml_str += f"{canonical_base},{prob_str};"
    else:
        mm_str, ml_str = None, None

    return seq, qual, mm_str, ml_str


def splits_to_intervals(positions, length):
    """Split to intervals."""
    if len(positions) == 0:
        return [(0, length)]
    prefix, suffix = [], []
    if positions[0] != 0:
        prefix = [0]
    if positions[-1] != length:
        suffix = [length]
    breaks = prefix + positions + suffix
    return [(start, end) for start, end in zip(breaks[:-1], breaks[1:])]


def get_enzyme(enzyme):
    """Get enzyme."""
    enz = getattr(Restriction, enzyme, None)
    if enz is None:
        raise ValueError(f"Enzyme not found: {enzyme}")
    if enz.cut_twice():
        raise NotImplementedError(
            f"Enzyme cuts twice, not currently supported: {enzyme}"
        )
    return enz


def get_concatemer_seqs(input_file, enzyme, remove_tags):
    """Digest concatemers in to unaligned monomers.

    :param input_file: The input BAM or Fastq file.
    :param enzyme: Name of the digestion enzyme used in sample preperation
    :param remove_tags: Comma seperated list of additional
                        tags to remove from file

    Concatemers are split with chosen digestion enzyme in to monomers.
    Monomers are tagged with unique monomer id and concatemer info.
    """
    logger.info(f"Digesting unaligned sequences from {input_file}")
    n_concatemers = 0
    n_monomers = 0
    enzyme = get_enzyme(enzyme)
    tags_remove = {"Ml", "ML", "Mm", "MM", "mv"}
    if remove_tags:
        tags_remove.update(set(remove_tags))
    for align in input_file.fetch(until_eof=True):
        n_concatemers += 1
        concatemer_id = align.query_name
        cut_points = [x - 1 for x in enzyme.search(Seq(align.query_sequence))]
        read_length = len(align.query_sequence)
        num_digits = len(str(read_length))
        intervals = splits_to_intervals(cut_points, read_length)
        num_intervals = len(intervals)
        for idx, (start, end) in enumerate(intervals):
            n_monomers += 1
            read = copy.copy(align)
            seq, qual, mm_str, ml_str = get_subread(
                align=align, start=start, end=end)
            for item in tags_remove:
                read.set_tag(item, None)
            read.query_sequence = seq
            read.query_qualities = qual
            # lexograpically sortable monomer ID
            read.query_name = \
                f"{concatemer_id}:{start:0{num_digits}d}:{end:0{num_digits}d}"
            utils.MonomerData.set_monomer_data(
                read, start, end, read_length, idx, num_intervals)
            read.set_tag(utils.CONCATEMER_ID_TAG, concatemer_id, "Z")
            read.set_tag("MM", mm_str)
            read.set_tag("ML", ml_str)
            yield read

    logger.info(f"Found {n_monomers} monomers in {n_concatemers} concatemers.")



"""Test functions from digest."""
import pytest

from pore_c_py import digest


@pytest.mark.parametrize(
    "length,positions,expected",
    [
        (10, [5], [(0, 5), (5, 10)]),
        (10, [0, 5], [(0, 5), (5, 10)]),
        (10, [0, 5, 10], [(0, 5), (5, 10)]),
        (10, [], [(0, 10)]),
        (10, [0], [(0, 10)]),
        (10, [0, 10], [(0, 10)]),
    ],
)
def test_splits_to_intervals(length, positions, expected):
    """Test splits to intervals."""
    intervals = digest.splits_to_intervals(positions, length)
    assert intervals == expected



"""Annotation of monomer alignments."""
from itertools import groupby

from pore_c_py import utils

logger = utils.get_named_logger("AnntateAln")


def get_walk_component(monomer):
    """Get monomer tag."""
    monomer_data = utils.MonomerData.from_pysam(monomer)
    if monomer.is_unmapped:
        monomer_name = (
            f"*:"
            f"{monomer_data.start}-{monomer_data.end}")
    else:
        orientation = "+-"[monomer.is_reverse]
        monomer_name = (
            f"{monomer.reference_name}:{orientation}:"
            f"{monomer.reference_start}-{monomer.reference_end}:"
            f"{monomer_data.start}-{monomer_data.end}")
    return monomer_name


def sort_by_category(alns, monomer_id):
    """Sort by category."""
    reordered = list()
    for aln in alns:
        # original claims this as "primary", thats not technically accurate
        category = 0
        if aln.is_secondary:
            category = 3
        elif aln.is_supplementary:
            category = 2
        elif aln.is_unmapped:
            category = 1
        reordered.append(list([aln, category]))
    # create a generator to be consistent with input
    reordered.sort(key=lambda a: a[1])
    if reordered[0][1] not in {0, 1}:
        logger.warning(
            f"Warning: best alignment for monomer: {monomer_id} "
            f"is secondary or supplementary")
    yield from (x[0] for x in reordered)


def annotate_alignments(bamfile):
    """Annotate (and filter) monomer alignments.

    :param bamfile: pysam.AlignmentFile
    :yields: the file header following by lists of alignments by concatemer.

    Alignments are grouped by concatemer (the input should be sorted by
    concatemer stored in the concatemer ID tag), and one alignment is selected
    per monomer. Alignments are annotated with a "walk", which simply
    enumerates the alignment coordinates of the monomers comprising the
    concatemer.
    """
    n_monomers = 0
    seen = set()
    for concat_id, aligns in groupby(
            bamfile.fetch(until_eof=True),
            lambda x: x.get_tag(utils.CONCATEMER_ID_TAG)):
        if concat_id in seen:
            logger.warning(
                f"Input file appears unsorted, found {concat_id} "
                "more than once.")
        seen.add(concat_id)
        walk = list()
        for monomer_id, alns in groupby(aligns, lambda x: x.query_name):
            n_monomers += 1
            sorted_aligns = sort_by_category(alns, monomer_id)
            aln = next(sorted_aligns)
            walk.append(aln)
        expected_monomers = utils.MonomerData.from_pysam(
            walk[-1]).subread_total
        concat_id = walk[-1].get_tag(utils.CONCATEMER_ID_TAG)
        if expected_monomers != len(walk):
            logger.warning(
                f"Expected to see {expected_monomers} alignments for "
                f"concatemer {concat_id}, found {len(walk)}"
            )
        walk_tag = ";".join(get_walk_component(aln) for aln in walk)
        for aln in walk:
            aln.set_tag(utils.WALK_TAG, walk_tag)
        yield walk
    logger.info(f"Found {n_monomers} monomers in {len(seen)} concatemers.")
